
/*
 *  source - $Id: machine.c,v 1.7 1997/06/13 01:12:49 qian Exp $
 *
 *  machine dependent functions
 */

#include <linux/config.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/head.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/smp.h>
#ifdef CONFIG_BLK_DEV_INITRD
#include <linux/blk.h>
#endif

#define printf	printk

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/pgtable.h>
#include <asm/dma.h>

#include <asm/memguard_machdep.h>
#include <memguard/macros.h>
#include <memguard/types.h>

#include <asm/memguard.h>

#include <memguard/memguard_internal.h>
#include <memguard/internal.h>


/*********************************************************************
variables (public)
*********************************************************************/


/*********************************************************************
variables (private)
*********************************************************************/

/*
 * Variables  : mgFreeAddr
 * Description: current free address
 */
/* current free address */
static u_int mgFreeAddr;


/*********************************************************************
function prototypes (private)
*********************************************************************/

static pte_t *mg_getpte(int pageno);
static pte_t *mg_getpte0(int pageno);


/*********************************************************************
functions (public)
*********************************************************************/

/*
 * Function   : mgBeforeInit
 * Description: machine-dependent work before initialization
 * Parameters : memory_start - the starting address of the free memory
 * Returns    : the new starting address of the free memory (u_int)
 * Plan       : 1. set current free address
 *              2. return the new starting address of the free memory
 * Note       : none
 */
u_int
mgBeforeInit(u_int memory_start)
{
	/*
	 * 1. set current free address
	 */
	mgFreeAddr = PAGE_ALIGN(memory_start);

	/*
	 * 2. return the new starting address of the free memory
	 */
	return mgFreeAddr;
}

/*
 * Function   : mgAfterInit
 * Description: machine-dependent work after initialization
 * Parameters : memory_start - the starting address of the free memory
 *              memory_end   - the ending address of the physical memory
 * Returns    : the new starting address of the free memory (u_int)
 * Plan       : 1. add dynamic memory from 0x1000 (4KB) to 0x9F000 (636KB)
 *              2. add kernel data + kernel bss region + dynamic memory
 *              3. calculate new memory address
 *              4. return new memory address
 * Note       : none
 */
u_int
mgAfterInit(u_int memory_start, u_int memory_end)
{
	u_int memory_start_new;

	extern char _etext;
	/* we're going to guard almost all the physical memory.
	   NOTE: if using __mgRecordCounter() in "arch/i386/kernel/ptrace.c"
	   to measure performance.  Set the guarded region to data+bss, or
	   data+bss+runtimeDS.  Do not include dynamicMemory.
	*/
	extern char _end;

	/*
	 * 1. add dynamic memory from 0x1000 (4KB) to 0x9F000 (640KB - 0x4KB)
	 */
	if (mgAddRegion((caddr_t)0x1000, 0x9F000 - 0x1000)
	    == MG_ADDREG_OK) {
printk("MEMGUARD: mgAfterInit(%x, %x), mgAddRegion(%#x, %#x) succeeded\n",
	memory_start, memory_end, 0x1000, 0x9F000 - 0x1000);
	} else {
		/* add region failed, report error */
printk("MEMGUARD: mgAfterInit(%x, %x), mgAddRegion(%#x, %#x) failed\n",
	memory_start, memory_end, 0x1000, 0x9F000 - 0x1000);
	}

	/*
	 * 2. add kernel data + kernel bss region + dynamic memory
	 */
	if (mgAddRegion((caddr_t) &_etext, memory_end - (int) &_etext)
	    == MG_ADDREG_OK) {
printk("MEMGUARD: mgAfterInit(%x, %x), mgAddRegion(%#x, %#x) succeeded\n",
	memory_start, memory_end, (int) &_etext, memory_end - (int) &_etext);
	} else {
		/* add region failed, report error */
printk("MEMGUARD: mgAfterInit(%x, %x), mgAddRegion(%#x, %#x) failed\n",
	memory_start, memory_end, (int) &_etext, memory_end - (int) &_etext);
	}

	/*
	 * 3. calculate new memory address
	 */
	memory_start_new = PAGE_ALIGN(mgFreeAddr);

	/*
	 * 4. return new memory address
	 */
	return memory_start_new;
}

/*
 * Function   : mgAllocate
 * Description: allocate memory space
 * Parameters : nunit  - the number of units
 *              szunit - the size of unit
 * Returns    : pointer to the allocated space (char *)
 *              != NULL	pointer to the allocated space
 *              == NULL	allocate failed
 * Plan       : 1. align current free address to word (4 bytes)
 *              2. set new current free address
 *              3. return allocated space pointer
 * Note       : called by mgInit() and mgAddRegion()/mg_addregion() only
 */
char *
mgAllocate(int nunit, int szunit)
{
    extern u_int mgFreeAddr;

printk("MEMGUARD: mgAllocate(%d, %d), old free address = %x\n",
        nunit, szunit, mgFreeAddr);
    /*
     * 1. align current free address to word (4 bytes)
     */
    mgFreeAddr = ((mgFreeAddr + (4 - 1)) / 4) * 4;
printk("MEMGUARD: mgAllocate(%d, %d), aligned free address = %x\n",
        nunit, szunit, mgFreeAddr);

    /*
     * 2. set new current free address
     */
    mgFreeAddr += nunit * szunit;
printk("MEMGUARD: mgAllocate(%d, %d), new free address = %x\n",
        nunit, szunit, mgFreeAddr);

    /*
     * 3. return allocated space pointer
     */
    return((char *)mgFreeAddr);
}

/*
 * Function   : mgGetProt
 * Description: get protection (including MM and TLB)
 * Parameters : pageno - physical page number
 *              p_prot - pointer to protection space for return
 * Returns    : none
 * Plan       : 1. get page table entry
 *              2. get original protection
 * Note       : none
 */
void
mgGetProt(int pageno, unsigned char *p_prot)
{
	pte_t *pgtable;

#ifdef MG_DEBUG
	printk("enter mgGetProt()\n");
#endif
	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte(pageno)) == NULL) {
		mgReportImpossible("mgGetProt(): mg_getpte()");
		return;
	}

	/*
	 * 2. get original protection
	 */
#ifdef MG_DEBUG
	printk("mgGetProt(pageno=%x): *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(*pgtable));
#endif
	*p_prot = (pte_val(*pgtable) & _PAGE_RW) ? MG_RDWR : MG_RDONLY;
#ifdef MG_DEBUG
	printk("exit mgGetProt()\n");
#endif
}

/*
 * Function   : mgGetInitProt
 * Description: get initial protection (including MM and TLB)
 * Parameters : pageno - physical page number
 *              p_prot - pointer to protection space for return
 * Returns    : none
 * Plan       :	1. get page table entry
 *              2. get original protection
 * Note       : none
 */
void
mgGetInitProt(int pageno, unsigned char *p_prot)
{
	pte_t *pgtable;

#ifdef MG_DEBUG
	printk("enter mgGetInitProt()\n");
#endif
	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte0(pageno)) == NULL) {
		mgReportImpossible("mgGetProt(): mg_getpte()");
		return;
	}

	/*
	 * 2. get original protection
	 */
#ifdef MG_DEBUG
	printk("mgGetInitProt(pageno=%x): *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(*pgtable));
#endif
	*p_prot = (pte_val(*pgtable) & _PAGE_RW) ? MG_RDWR : MG_RDONLY;
#ifdef MG_DEBUG
	printk("exit mgGetInitProt()\n");
#endif
}

/*
 * Function   : mgSetProt
 * Description: set protection (including MM and TLB)
 * Parameters : pageno - physical page number
 *              prot   - page protection
 * Returns    : none
 * Plan       : 1. get page table entry
 *              2. set protection
 *              3. flush the TLB entry
 * Note       : none
 */
void
mgSetProt(int pageno, unsigned char prot)
{
	pte_t *pgtable, entry;
	unsigned long page;

#ifdef MG_DEBUG
	printk("enter mgSetProt()\n");
#endif
	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte(pageno)) == NULL) {
		mgReportImpossible("mgSetProt(): mg_getpte()");
		return;
	}

	/*
	 * 2. set protection
	 */
	entry = *pgtable;
#ifdef MG_DEBUG
	printk("mgSetProt(pageno=%x): before *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif
	pte_val(entry) = (pte_val(entry) & ~_PAGE_RW) | (prot ? _PAGE_RW : 0);
#ifdef MG_DEBUG
	printk("mgSetProt(pageno=%x): after  *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif

	set_pte(pgtable, entry);
/* incorrect because protection bits include not only read-write bit
	printk("mgSetProt(pageno=%x): before *%x=%08lx\n",
		pageno, pgtable, *pgtable);
	set_pte(pgtable, pte_modify(*pgtable, prot ? _PAGE_RW : 0));
	printk("mgSetProt(pageno=%x): after  *%x=%08lx\n",
		pageno, pgtable, *pgtable);
*/

	/*
	 * 3. flush the TLB entry
	 */
	page = pageno << PAGE_SHIFT;
#ifdef MG_DEBUG
	printk("mgSetProt(): before __flush_tlb_one(%#x)\n", (int)page);
#endif
        __flush_tlb_one(page);
#ifdef MG_DEBUG
	printk("exit mgSetProt()\n");
#endif
}

/*
 * Function   : mgSetReadOnly
 * Description: set machine-dependent read-only protection
 *              (including MM and TLB)
 * Parameters : pageno - physical page number
 *              p_prot - pointer to protection space for return
 * Returns    : none
 * Plan       : 1. get page table entry
 *              2. get original protection
 *              3. set new protection
 *              4. flush the TLB entry
 * Note       : none
 */
void
mgSetReadOnly(int pageno, unsigned char *p_prot)
{
	pte_t *pgtable, entry;
	unsigned long page;

#ifdef MG_DEBUG
	printk("enter mgSetReadOnly()\n");
#endif
	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte(pageno)) == NULL) {
		mgReportImpossible("mgSetReadOnly(): mg_getpte()");
		return;
	}

	/*
	 * 2. get original protection
	 */
	*p_prot = (pte_val(*pgtable) & _PAGE_RW) ? MG_RDWR : MG_RDONLY;

	/*
	 * 3. set new protection
	 */
	if (! *p_prot) {
		/* read-only, do nothing, return */
		return;
	}
	/* read-write, set to read-only protection */

	entry = *pgtable;
#ifdef MG_DEBUG
	printk("mgSetReadOnly(pageno=%x): before *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif
	pte_val(entry) = pte_val(entry) & ~_PAGE_RW;
#ifdef MG_DEBUG
	printk("mgSetReadOnly(pageno=%x): after  *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif

	set_pte(pgtable, entry);
/* incorrect because protection bits include not only read-write bit
	printk("mgSetReadOnly(pageno=%x): before *%x=%08lx\n",
		pageno, pgtable, *pgtable);
	set_pte(pgtable, pte_modify(*pgtable, 0);
	printk("mgSetReadOnly(pageno=%x): after  *%x=%08lx\n",
		pageno, pgtable, *pgtable);
*/

	/*
	 * 4. flush the TLB entry
	 */
	page = pageno << PAGE_SHIFT;
#ifdef MG_DEBUG
	printk("mgSetReadOnly(): before __flush_tlb_one(%#x)\n", (int)page);
#endif
        __flush_tlb_one(page);
#ifdef MG_DEBUG
	printk("exit mgSetReadOnly()\n");
#endif
}

/*
 * Function   : mgSetReadWrite
 * Description: set machine-dependent read-write protection
 *              (including MM and TLB)
 * Parameters : pageno - physical page number
 *              p_prot - pointer to protection space for return
 * Returns    : none
 * Plan       : 1. get page table entry
 *              2. get original protection
 *              3. set new protection
 *              4. flush the TLB entry
 * Note       : none
 */
void
mgSetReadWrite(int pageno, unsigned char *p_prot)
{
	pte_t *pgtable, entry;
	unsigned long page;

#ifdef MG_DEBUG
	printk("enter mgSetReadWrite()\n");
#endif
	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte(pageno)) == NULL) {
		mgReportImpossible("mgSetReadWrite(): mg_getpte()");
		return;
	}

	/*
	 * 2. get original protection
	 */
	*p_prot = (pte_val(*pgtable) & _PAGE_RW) ? MG_RDWR : MG_RDONLY;

	/*
	 * 3. set new protection
	 */
	if (*p_prot) {
		/* read-write, do nothing, return */
		return;
	}
	/* read-only, set to read-write protection */

	entry = *pgtable;
#ifdef MG_DEBUG
	printk("mgSetReadWrite(pageno=%x): before *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif
	pte_val(entry) = pte_val(entry) | _PAGE_RW;
#ifdef MG_DEBUG
	printk("mgSetReadWrite(pageno=%x): after  *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(entry));
#endif

	set_pte(pgtable, entry);
/* incorrect because protection bits include not only read-write bit
	printk("mgSetReadWrite(pageno=%x): before *%x=%08lx\n",
		pageno, pgtable, *pgtable);
	set_pte(pgtable, pte_modify(*pgtable, _PAGE_RW);
	printk("mgSetReadWrite(pageno=%x): after  *%x=%08lx\n",
		pageno, pgtable, *pgtable);
*/

	/*
	 * 4. flush the TLB entry
	 */
	page = pageno << PAGE_SHIFT;
#ifdef MG_DEBUG
	printk("mgSetReadWrite(): before __flush_tlb_one(%#x)\n", (int)page);
#endif
        __flush_tlb_one(page);
#ifdef MG_DEBUG
	printk("exit mgSetReadWrite()\n");
#endif
}

/*
 * Function   : mgRead
 * Description: read in a page
 * Parameters : content- read content
 *              pageno - physical page number
 *              offset - read offset in the page
 *              size   - read size in the page
 * Returns    : read length (int)
 * Plan       : 1. read page
 *              2. return read length
 * Note       : none
 */
int
mgRead(char *content, int pageno, u_short offset, u_short size)
{
    /*
     * 1. read page
     */
	memcpy(content, (char *)((pageno << PAGE_SHIFT) + offset), size);

    /*
     * 2. return read length
     */
    return(size);
}

/*
 * Function   : mgWrite
 * Description: write in a page
 * Parameters : content- write content
 *              pageno - physical page number
 *              offset - write offset in the page
 *              size   - write size in the page
 * Returns    : write length (int)
 * Plan       : 1. write page
 *              2. return written length
 * Note       : none
 */
int
mgWrite(char *content, int pageno, u_short offset, u_short size)
{
    /*
     * 1. write page
     */
	if (! mgEnabled) {
		memcpy((char *)((pageno << PAGE_SHIFT) + offset),
			content, size);
	} else {
		/* interrupts disabled in mgSet()
		flags_t flags;
		*/
		unsigned char prot_old;

		/*
		__mgCli(flags);
		*/

		mgSetReadWrite(pageno, &prot_old);
		memcpy((char *)((pageno << PAGE_SHIFT) + offset),
			content, size);
		mgSetProt(pageno, prot_old);

		/*
		__mgSti(flags);
		*/
	}

    /*
     * 2. return written length
     */
    return(size);
}

/*
 * Function   : mgAnalyze
 * Description: analyze the current instruction, return the writing length
 * Parameters : eip - the content of EIP register
 * Returns    : the writing length (int)
 * Plan       : 1. return the writing length
 * Note       : none
 */
int
mgAnalyze(long eip)
{
	/*
	 * 1. return the writing length
	 */
	return 4;
}

#define MG_FAULT_PAGES	10

int mgFaultPages = 0;	/* mgFaultPages is vulnerable to interrupts because
				interrupt handlers may write to guarded pages */

static pte_t * mgFaultPteArray[MG_FAULT_PAGES];
static int mgFaultPgnArray[MG_FAULT_PAGES];
static int mgIntEnabled;
static int mgSSEnabled;
/*
 * NOTE: MG_TEST cannot be defined along with MG_DEBUG, MG_PERF_INTERRUPT, ...
 *	and should be defined only with MG_TEST in "arch/i386/kernel/ptrace.c"
 * NOTE: MG_PERF_INTERRUPT cannot be defined along with MG_DEBUG, MG_TEST, ...
 *	and should be defined only with MG_PERF_INTERRUPT in
 *	"arch/i386/kernel/ptrace.c"
 */
#ifndef MG_DEBUG
/*
#define MG_TEST			check "arch/i386/kernel/ptrace.c"
#define MG_TEST_REPORT_FREQ	1
*/
/*
#define MG_PERF_INTERRUPT	check "arch/i386/kernel/ptrace.c"
*/
#endif
#ifdef MG_PERF_INTERRUPT
static int mgStartCounterLow, mgStartCounterHigh;
static int mgEndCounterLow, mgEndCounterHigh;
#endif 
/*
 * Function   : mgSetEnvironment
 * Description: change page protection to read-write, set to signal-step mode
 * Parameters : regs   - the pointer to the structure of the registers
 *              pageno - physical page number
 * Returns    : none
 * Plan       : 1. get page table entry
 *              2. disable interrupt
 *              3. record information
 *              4. update stacked EFLAGS
 *              5. set protection
 *              6. flush the TLB entry
 * Note       : none
 */
void
mgSetEnvironment(struct pt_regs *regs, int pageno)
{
	pte_t *pgtable, entry;
	unsigned long page;

	/*
	 * 1. get page table entry
	 */
	if ((pgtable = mg_getpte(pageno)) == NULL) {
		mgReportImpossible("mgSetEnvironment(): mg_getpte()");
		return;
	}
/*
printk("Set set_pte: *%x = %08lx\n", (int)pgtable, pte_val(*pgtable));
*/

	/*
	 * 2. disable interrupt
	 */
	cli();	/* prevent an interrupt handler from triggering a MemGuard
			page protection fault */
#ifdef MG_PERF_INTERRUPT
	__mgRecordCounter(mgStartCounterLow, mgStartCounterHigh);
#endif
	/*
	 * 3. record information
	 */
	if (mgFaultPages >= MG_FAULT_PAGES) {
		mgReportImpossible(
			"mgSetEnvironment(): mgFaultPages >= MG_FAULT_PAGES");
		printk("mgFaultPages = %d\n", mgFaultPages);
		return;
	}
	if (mgFaultPages == 0) {
		mgIntEnabled = regs->eflags & 0x00000200;
		mgSSEnabled = regs->eflags & 0x00000100;
	}
	mgFaultPteArray[mgFaultPages] = pgtable;
	mgFaultPgnArray[mgFaultPages ++] = pageno;

	/*
	 * 4. update stacked EFLAGS
	 */
	regs->eflags &= ~0x00000200;
	regs->eflags |= 0x00000100;

	/*
	 * 5. set protection
	 */
#ifdef MG_DEBUG
	printk("mgSetEnvironment(pageno=%x): before *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(*pgtable));
#endif
	entry = *pgtable;
	pte_val(entry) |= _PAGE_RW;

	set_pte(pgtable, entry);
/* can also use pte_mkwrite() directly
	set_pte(pgtable, pte_mkwrite(*pgtable));
*/
#ifdef MG_DEBUG
	printk("mgSetEnvironment(pageno=%x): after  *%x=%08lx\n",
		pageno, (int)pgtable, pte_val(*pgtable));
#endif
/*
printk("Set after set_pte: *%x = %08lx\n", (int)pgtable, pte_val(*pgtable));
*/

	/*
	 * 6. flush the TLB entry
	 */
	page = pageno << PAGE_SHIFT;
#ifdef MG_DEBUG
	printk("mgSetEnvironment(): before __flush_tlb_one(%#x)\n", (int)page);
#endif
        __flush_tlb_one(page);
}

/*
 * Function   : mgResetEnvironment
 * Description: restore page protection and single-step mode
 * Parameters : regs   - the pointer to the structure of the registers
 * Returns    : if the single-step mode is formerly enabled (int)
 * Plan       : 1. restore page protection
 *              2. restore interrupt
 *              3. update stacked EFLAGS
 *              4. return if the single-step mode is formerly enabled
 * Note       : none
 */
int
mgResetEnvironment(struct pt_regs *regs)
{
	int i;

	/*
	 * 1. restore page protection
	 */
	for (i = mgFaultPages - 1; i >= 0; i --) {
		unsigned long page;
/* we'd better use pte_wrprotect()
		pte_t entry;

		entry = *mgFaultPteArray[i];
		pte_val(entry) &= ~_PAGE_RW;

		set_pte(mgFaultPteArray[i], entry);
*/
#ifdef MG_DEBUG
		printk("mgResetEnvironment(): pageno=%x, before *%x=%08lx\n",
			mgFaultPgnArray[i],
			(int)mgFaultPteArray[i], pte_val(*mgFaultPteArray[i]));
#endif
/*
printk("Reset set_pte: *%x = %08lx\n", (int)mgFaultPteArray[i], pte_val(*mgFaultPteArray[i]));
*/
		set_pte(mgFaultPteArray[i], pte_wrprotect(*mgFaultPteArray[i]));
/*
printk("Reset after set_pte: *%x = %08lx\n", (int)mgFaultPteArray[i], pte_val(*mgFaultPteArray[i]));
*/
#ifdef MG_DEBUG
		printk("mgResetEnvironment(): pageno=%x, after  *%x=%08lx\n",
			mgFaultPgnArray[i],
			(int)mgFaultPteArray[i], pte_val(*mgFaultPteArray[i]));
#endif
        	page = mgFaultPgnArray[i] << PAGE_SHIFT;
#ifdef MG_DEBUG
		printk("mgResetEnvironment(): before __flush_tlb_one(%#x)\n",
			(int)page);
#endif
        	__flush_tlb_one(page);
	}
	mgFaultPages = 0;

	/*
	 * 2. restore interrupt
	 */
	if (mgIntEnabled) {
#ifdef MG_PERF_INTERRUPT
		__mgRecordCounter(mgEndCounterLow, mgEndCounterHigh);
#endif
		sti();
	} else {
#ifdef MG_PERF_INTERRUPT
		__mgRecordCounter(mgEndCounterLow, mgEndCounterHigh);
#endif
	}
#ifdef MG_PERF_INTERRUPT
	__mgPrintCounter("Disable Interrupt",
		mgStartCounterLow, mgStartCounterHigh);
	__mgPrintCounter("Enable  Interrupt",
		mgEndCounterLow, mgEndCounterHigh);
#endif

	/*
	 * 3. update stacked EFLAGS
	 */
	if (mgIntEnabled) {
		regs->eflags |= 0x00000200;
	}
	if (! mgSSEnabled) {
		regs->eflags &= ~0x00000100;
	}

	/*
	 * 4. return if the single-step mode is formerly enabled
	 */
	return(mgSSEnabled);
}

static int mgPrevPgn;
static int mgPrevPages;				/* 0 <= mgPrevPages <=2 */
static unsigned char mgPrevProtArray[2];	/* at most two pages */

/*
 * Function   : mgBeforeSwitch
 * Description: set the current task structure to read-write
 * Parameters : prev - the current task pointer
 * Returns    : none
 * Plan       : 1. in case the current task is dead (I don't think so)
 *            : 2. set to read-write & store the original page protection(s)
 * Note       : none
 */
void
mgBeforeSwitch(struct task_struct *prev)
{
	int i, pageno;

	/*
	 * 1. in case the current task is dead (I don't think so)
	 */
	if (prev == NULL) {
		mgPrevPages = 0;
		return;
	}

	/*
	 * 2. set to read-write & store the original page protection(s)
	 */
	for (i = 0, mgPrevPgn = pageno = mgAddr2Pageno((int)prev);
	    pageno < mgAddr2Pageno((int)prev + sizeof(*prev) - 1) + 1;
	    i++, pageno ++) {
		mgSetReadWrite(pageno, &mgPrevProtArray[i]);
	}
	mgPrevPages = i;
}

/*
 * Function   : mgAfterSwitch
 * Description: restore the protection of the current task structure
 * Parameters : none
 * Returns    : none
 * Plan       : 1. restore the original page protection(s)
 * Note       : none
 */
void
mgAfterSwitch()
{
	int i, pageno;

	/*
	 * 1. restore the original page protection(s)
	 */
	for (i = 0, pageno = mgPrevPgn; i < mgPrevPages; i ++, pageno ++) {
		mgSetProt(pageno, mgPrevProtArray[i]);
	}
}

/*
 * Function   : mgReportImpossible
 * Description: report programmer's error
 * Parameters : msg - error message
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportImpossible(char *msg)
{
	printk("MEMGUARD (IMPOSSIBLE): %s\n", msg);
}

/*
 * Function   : mgReportFault
 * Description: report memguard fault
 * Parameters : vaddr      - writing virtual address
 *              length     - write length in the page
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportFault(caddr_t vaddr, u_short length)
{
#ifdef MG_TEST
{
    static int count = 0;

    if ((++ count) >= MG_TEST_REPORT_FREQ) {
	count = 0;

        printk("MEMGUARD: mgFault(), vaddr=%#x length=%#x\n",
            (int)vaddr, (int)length);
    }
}
#else	/* MG_TEST */
    printk("MEMGUARD: mgFault(), vaddr=%#x length=%#x\n",
        (int)vaddr, (int)length);
#endif	/* MG_TEST */
}

/*
 * Function   : mgReportViolate
 * Description: report invariant violation
 * Parameters : pinvariant - pointer to invariant violated
 *              vaddr      - writing virtual address
 *              offset     - write offset in the page
 *              length     - write length in the page
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportViolate(mginvariant_t *pinvariant,
		  caddr_t vaddr, u_short offset, u_short length)
{
#ifdef MG_TEST
{
    static int count = 0;

    if ((++ count) >= MG_TEST_REPORT_FREQ) {
	count = 0;

        printk("MEMGUARD (violate): vaddr=%#x offset=%#x length=%#x\n",
            (int)vaddr, (int)offset, (int)length);
        printk("\twrites: name=\"%.*s\" offset=%#x length=%#x size=%#x\n",
            MG_SZNAME, pinvariant->gi_name, pinvariant->gi_offset,
	    pinvariant->gi_length, pinvariant->gi_size);
    }
}
#else	/* MG_TEST */
    printk("MEMGUARD (violate): vaddr=%#x offset=%#x length=%#x\n",
        (int)vaddr, (int)offset, (int)length);
    printk("\twrites: name=\"%.*s\" offset=%#x length=%#x size=%#x\n",
        MG_SZNAME, pinvariant->gi_name, pinvariant->gi_offset,
	pinvariant->gi_length, pinvariant->gi_size);
#endif	/* MG_TEST */
}

/*
 * Function   : mgReportConflict
 * Description: report invariant conflict
 * Parameters : pinvariant - pointer to invariant conflicted
 *              name       - new invariant name
 *              vaddr      - new invariant virtual address
 *              size       - new invariant size
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportConflict(mginvariant_t *pinvariant,
		 char *name, caddr_t vaddr, u_int size)
{
    printk("MEMGUARD (conflict): name=\"%.*s\" vaddr=%#x size=%#x\n",
        MG_SZNAME, name, (int)vaddr, (int)size);
    printk("\toverlaps: name=\"%.*s\" offset=%#x length=%#x size=%#x\n",
        MG_SZNAME, pinvariant->gi_name, pinvariant->gi_offset,
	pinvariant->gi_length, pinvariant->gi_size);
}

/*
 * Function   : mgReportSize
 * Description: report invariant size error
 * Parameters : pinvariant - pointer to invariant with different size
 *              vaddr      - searching virtual address
 *              size       - searching size
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportSize(mginvariant_t *pinvariant, caddr_t vaddr, u_int size)
{
    printk("MEMGUARD (invalid size): vaddr=%#x size=%#x\n",
	(int)vaddr, (int)size);
    printk("\toriginal: name=\"%.*s\" offset=%#x length=%#x size=%#x\n",
        MG_SZNAME, pinvariant->gi_name, pinvariant->gi_offset,
        pinvariant->gi_length, pinvariant->gi_size);
}

/*
 * Function   : mgReportNotfound
 * Description: report invariant not found
 * Parameters : pinvariant - pointer to invariant with different size
 *              vaddr      - searching virtual address
 *              size       - searching size
 * Returns    : none
 * Plan       :
 * Note       : none
 */
void
mgReportNotfound(mginvariant_t *pinvariant, caddr_t vaddr, u_int size)
{
    printk("MEMGUARD (non-exist): vaddr=%#x size=%#x\n", (int)vaddr, (int)size);
}


/*********************************************************************
functions (private)
*********************************************************************/

/*
 * Function   : mg_getpte
 * Description: get page table entry
 * Parameters : pageno - physical page number
 * Returns    : pointer to the page table entry (pte_t *)
 *              != NULL	pointer to the page table entry
 *              == NULL	failed
 * Plan       : 1. calculate linear address
 *		2. get page directory entry
 *		3. get middle page directory entry
 *		4. get page table entry
 *		5. return page table entry
 * Note       : none
 */
static pte_t *
mg_getpte(int pageno)
{
	pgd_t * pgdir;
	pmd_t * pgmiddle;
	pte_t * pgtable;
	unsigned long page;

        /*
         * 1. calculate linear address
         */
	page = (pageno << PAGE_SHIFT) + 0xC0000000;
#ifdef MG_DEBUG
	printk("mg_getpte(%x): page = %x\n", pageno, (int)page);
#endif

	/*
	 * 2. get page directory entry
	 */
	pgdir = swapper_pg_dir + (page >> PGDIR_SHIFT);
#ifdef MG_DEBUG
	printk("mg_getpte(%x): pgdir = %x\n", pageno, (int)pgdir);
#endif
	if (pgd_none(*pgdir)) {
		printk("mg_getpte(%x): no kernel pgdir, impossible\n", pageno);
		return NULL;
	}
	if (pgd_bad(*pgdir)) {
		printk("mg_getpte(%x): bad kernel pgdir, impossible\n", pageno);
		return NULL;
	}

	/*
	 * 3. get middle page directory entry
	 */
	pgmiddle = pmd_offset(pgdir, page);
#ifdef MG_DEBUG
	printk("mg_getpte(%x): pgmiddle = %x\n", pageno, (int)pgmiddle);
#endif
	if (pmd_none(*pgmiddle)) {
		printk("mg_getpte(%x): no kernel pgmiddle, impossible\n",
			pageno);
		return NULL;
	}
	if (pmd_bad(*pgmiddle)) {
		printk("TEST pmd_val(pmd) & ~PAGE_MASK) != _PAGE_TABLE ");
		printk("%x & ~%x != %x\n", (int)pmd_val(*pgmiddle),
			(int)PAGE_MASK, (int)_PAGE_TABLE);
		printk("TEST pmd_val(pmd) > high_memory ");
		printk("%x > %x\n", (int)pmd_val(*pgmiddle), (int)high_memory);
		printk("mg_getpte(%x): bad kernel pgmiddle, impossible\n",
			pageno);
		return NULL;
	}

	/*
	 * 4. get page table entry
	 */
	pgtable = pte_offset(pgmiddle, page);

	/*
	 * 5. return page table entry
	 */
	return pgtable;
}

/*
 * Function   : mg_getpte0
 * Description: get page table entry without boundary checking
 * Parameters : pageno - physical page number
 * Returns    : pointer to the page table entry (pte_t *)
 *              != NULL	pointer to the page table entry
 *              == NULL	failed
 * Plan       : 1. calculate linear address
 *		2. get page directory entry
 *		3. get middle page directory entry
 *		4. get page table entry
 *		5. return page table entry
 * Note       : none
 */
static pte_t *
mg_getpte0(int pageno)
{
	pgd_t * pgdir;
	pmd_t * pgmiddle;
	pte_t * pgtable;
	unsigned long page;

        /*
         * 1. calculate linear address
         */
	page = (pageno << PAGE_SHIFT) + 0xc0000000;
#ifdef MG_DEBUG
	printk("mg_getpte0(%x): page = %x\n", pageno, (int)page);
#endif

	/*
	 * 2. get page directory entry
	 */
	pgdir = swapper_pg_dir + (page >> PGDIR_SHIFT);
#ifdef MG_DEBUG
	printk("mg_getpte0(%x): pgdir = %x\n", pageno, (int)pgdir);
#endif
	if (pgd_none(*pgdir)) {
		printk("mg_getpte0(%x): no kernel pgdir, impossible\n", pageno);
		return NULL;
	}
	if (pgd_bad(*pgdir)) {
		printk("mg_getpte0(%x): bad kernel pgdir, impossible\n", pageno);
		return NULL;
	}

	/*
	 * 3. get middle page directory entry
	 */
	pgmiddle = pmd_offset(pgdir, page);
#ifdef MG_DEBUG
	printk("mg_getpte0(%x): pgmiddle = %x\n", pageno, (int)pgmiddle);
#endif
	if (pmd_none(*pgmiddle)) {
		printk("mg_getpte0(%x): no kernel pgmiddle, impossible\n",
			pageno);
		return NULL;
	}
	if ((pmd_val(*pgmiddle) & ~PAGE_MASK) != _PAGE_TABLE) {
		printk("TEST pmd_val(pmd) & ~PAGE_MASK) != _PAGE_TABLE ");
		printk("%x & ~%x != %x\n", (int)pmd_val(*pgmiddle),
			(int)PAGE_MASK, (int)_PAGE_TABLE);
		printk("mg_getpte0(%x): bad kernel pgmiddle, impossible\n",
			pageno);
		return NULL;
	}

	/*
	 * 4. get page table entry
	 */
	pgtable = pte_offset(pgmiddle, page);

	/*
	 * 5. return page table entry
	 */
	return pgtable;
}

